# -*- coding: utf-8 -*-
import boto3


class Folder:
    def __init__(self, prefix, files, subdirectories):
        self.prefix = prefix
        self.files = files
        self.subdirectories = subdirectories

    def get_index_hash(self):
        for file in self.files:
            if file["Key"] == f"{self.prefix}index.html":
                return file["ETag"].replace('"', "")
        return None


class S3:
    def __init__(self, bucket_name):
        self.s3_client = boto3.client("s3")
        self.bucket_name = bucket_name

    def fetch_folder_content(self, folder_key):
        paginator = self.s3_client.get_paginator("list_objects_v2")
        page_iterator = paginator.paginate(
            Bucket=self.bucket_name, Prefix=folder_key, Delimiter="/"
        )
        files = []
        subdirectories = []
        for page in page_iterator:
            files.extend(page.get("Contents", []))
            subdirectories.extend(page.get("CommonPrefixes", []))

        subdirectories = list(map(lambda data: data["Prefix"], subdirectories))

        return Folder(prefix=folder_key, subdirectories=subdirectories, files=files)

    def put_object(self, body, key):
        self.s3_client.put_object(
            Body=body,
            Bucket=self.bucket_name,
            CacheControl="max-age=0",
            ContentType="text/html",
            Key=key,
        )



# -*- coding: utf-8 -*-
import hashlib
import logging
from collections import deque
from concurrent.futures.thread import ThreadPoolExecutor

import boto3
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

from .index import Index
from .s3 import S3


class BucketDirGenerator:
    def __init__(
        self,
        bucket_name,
        site_name,
        logger=None,
    ):
        self.logger = logger or logging.getLogger("bucket_dir")
        self.template_environment = Environment(
            loader=PackageLoader("bucket_dir", "templates"),
            autoescape=select_autoescape(["html"]),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.site_name = site_name
        self.s3_gateway = S3(bucket_name=bucket_name)

    @staticmethod
    def generate_ascending_prefixes(directory_key):
        parts = directory_key.split("/")
        parts = list(filter(len, parts))
        paths = []
        for index in range(len(parts)):
            level = "/".join(parts[:index])
            if (not level.endswith("/")) and (level != ""):
                level += "/"
            paths.append(level)
        paths.reverse()
        return paths

    def generate(self, exclude_objects=None, single_threaded=False, target_path=""):
        if target_path.startswith("/"):
            target_path = target_path[1:]

        if not target_path.endswith("/"):
            last_slash = target_path.rfind("/")
            target_path = target_path[: last_slash + 1]

        max_workers = 1 if single_threaded else None

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            self.logger.info(f"Running {executor._max_workers} worker threads")
            futures = deque([])
            futures.append(
                executor.submit(
                    self.render_and_upload_index_document_to_s3,
                    target_path,
                    exclude_objects,
                    executor,
                )
            )

            for prefix in self.generate_ascending_prefixes(target_path):
                futures.append(
                    executor.submit(
                        self.render_and_upload_index_document_to_s3,
                        prefix,
                        exclude_objects,
                    )
                )

            self.wait_for_all_futures_recursively(futures)
            self.logger.info(f"Finished generation")

    def wait_for_all_futures_recursively(self, futures):
        self.logger.debug(f"waiting for all futures to finish")
        while len(futures) > 0:
            sub_futures_array = futures.popleft().result()
            futures.extend(sub_futures_array)

    def render_and_upload_index_document_to_s3(self, prefix, extra_excluded_items, executor=None):
        self.logger.debug(f"Rendering index for {prefix}.")
        folder = self.s3_gateway.fetch_folder_content(prefix)

        futures = []
        if executor is not None:
            for subdirectory in folder.subdirectories:
                futures.append(
                    executor.submit(
                        self.render_and_upload_index_document_to_s3,
                        subdirectory,
                        extra_excluded_items,
                        executor,
                    )
                )

        index = Index(prefix, extra_excluded_items)
        index.items.extend(folder.files)
        index.folders.extend(folder.subdirectories)

        index_document = index.render(
            site_name=self.site_name, template_environment=self.template_environment
        ).encode("utf-8")

        key = f"{prefix}index.html"

        new_hash = hashlib.md5(  # nosec # skip bandit check as this is not used for encryption
            index_document
        ).hexdigest()
        old_hash = folder.get_index_hash()
        self.logger.debug(f"{key} comparing existing hash: {old_hash} to new hash: {new_hash}")
        if old_hash == new_hash:
            self.logger.info(f"Skipping unchanged index for {key}")
        else:
            self.logger.info(f"Uploading index for {key}")
            self.s3_gateway.put_object(
                body=index_document,
                key=key,
            )

        return futures



# -*- coding: utf-8 -*-
import urllib

import humanize


class Index:
    def __init__(self, path, extra_excluded_items=None):
        self.path = path
        self.items = []
        self.folders = []
        self.excluded_items = ["favicon.ico", "index.html"]
        if extra_excluded_items:
            self.excluded_items.extend(extra_excluded_items)

    def should_exclude(self, file_name):
        return file_name in self.excluded_items

    def render(self, site_name, template_environment):
        template = template_environment.get_template("index.html.j2")
        folder_lengths = list(map(len, self.folders))
        file_lengths = [len(item["Key"]) for item in self.items]
        longest_name_length = max(folder_lengths + file_lengths + [5])

        longest_modified_length = max(
            [len(str(item["LastModified"])) for item in self.items] + [14]
        )

        column_padding = 3
        table_headers = f"{'Name'.ljust(longest_name_length + column_padding)}{'Last modified'.ljust(longest_modified_length + column_padding)}Size"

        index_items = []
        for item in self.items:
            file_name = item["Key"].split("/")[-1]

            if self.should_exclude(file_name):
                continue

            index_items.append(
                self.file_index_item(
                    column_padding, file_name, item, longest_modified_length, longest_name_length
                )
            )

        for folder in self.folders:
            folder_name = folder.split("/")[-2] + "/"
            index_items.append(
                self.folder_index_item(
                    column_padding,
                    folder,
                    folder_name,
                    longest_modified_length,
                    longest_name_length,
                )
            )

        return template.render(
            page_name=f"{site_name}/{self.path}",
            table_headers=table_headers,
            index_items=index_items,
            not_root=self.path != "",
        )

    def folder_index_item(
        self, column_padding, folder, folder_name, longest_modified_length, longest_name_length
    ):
        return {
            "encoded_name": urllib.parse.quote(folder_name),
            "name": folder_name,
            "name_padding": " ".ljust(longest_name_length - len(folder) + column_padding),
            "modified": "-".ljust(longest_modified_length + column_padding),
            "size": "-",
        }

    def file_index_item(
        self, column_padding, file_name, item, longest_modified_length, longest_name_length
    ):
        return {
            "encoded_name": urllib.parse.quote(file_name),
            "name": file_name,
            "name_padding": " ".ljust(longest_name_length - len(item["Key"]) + column_padding),
            "modified": item["LastModified"]
            .strftime("%d-%b-%Y %H:%M")
            .ljust(longest_modified_length + column_padding),
            "size": humanize.naturalsize(item["Size"]),
        }
